#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" LocationsReport conveniences """
import http
import logging

# pylint: disable=E0239, E0401, E0402, E1136, R0201, R0902, R0903, R0913, W0613
import requests
from enum import Enum
from marshmallow import Schema, fields, post_load
from marshmallow.decorators import validates_schema
from pathlib import Path
from pycapo import CapoConfig
from typing import NamedTuple, Optional

from productfetcher.interfaces import FetcherFactory
from .exceptions import FetchError, LocatorServiceException
from .interfaces import LocatedFile, LocationReport, Locator, FetchContext, FileFetcher
from .ngas import NgasConnection

logger = logging.getLogger("productfetcher.locations")
logger.setLevel(logging.INFO)


class Location(Enum):
    """
    Where the files live
    """

    DSOC = "DSOC"
    NAASC = "NAASC"
    UVA = "UVA"

    @property
    def lustre_prefix(self) -> str:
        """
        The expected Lustre path prefix for this location
        """
        return "/lustre/aoc" if self == self.DSOC else "/lustre/naasc"

    @staticmethod
    def in_location(path: Path) -> Optional["Location"]:
        """
        Given a path, return the location it appears to be in, or nothing,
        if you're out in space somewhere.

        :param path:  the path to test
        :return:  a Location, optionally
        """

        # look through the locations until we find one that seems to match
        for location in Location:
            if str(path.absolute()).startswith(location.lustre_prefix):
                return location

        return None


class Cluster(Enum):
    """
    Which cluster the files are on
    """

    DSOC = "DSOC"
    NAASC = "NAASC"


class NgasServer(NamedTuple):
    """
    Encapsulates the "server" section of a "files" item in a location report
    """

    server: str
    location: Location
    cluster: Cluster

    def can_direct_copy_to(self, dest: Path) -> bool:
        """
        True if we can direct copy to this destination.

        In principle, direct copy is possible if the file's NGAS server and the destination
        are in the same datacenter, but it also depends on the correct version of NGAS and
        the NGAS direct copy plugin, which is not supported by the NAASC cluster at the moment.
        So there is a test in here which will need to be removed eventually, to ensure that
        the DSOC cluster is accounted for.

        :param dest:  the destination to write to
        :return:      true if direct copy is possible
        """
        exec_site = CapoConfig().settings("edu.nrao.workspaces.ProductFetcherSettings").executionSite
        # Catch the case of local testing
        if "local" in exec_site:
            return False

        return self.cluster == Cluster.DSOC and self.location == Location.in_location(dest)

    def connect(self) -> NgasConnection:
        """
        Connect to this NGAS server and return an NgasConnection we can use to send requests to.

        :return: a connection to NGAS
        """
        return NgasConnection(self.server)

    def __str__(self):
        return f"NGAS {self.server} at {self.location}"


class NgasFile(LocatedFile):
    """Encapsulates all the information about an NGAS file in a locations report"""

    ngas_file_id: str
    subdirectory: str
    relative_path: str
    checksum: int
    checksum_type: str
    version: int
    size: int
    server: NgasServer
    _science_product_locator: str = None

    def __init__(
        self,
        ngas_file_id: str,
        subdirectory: str,
        relative_path: str,
        checksum: int,
        checksum_type: str,
        version: int,
        size: int,
        server: NgasServer,
    ):
        super().__init__()
        self.ngas_file_id = ngas_file_id
        self.subdirectory = subdirectory
        self.relative_path = relative_path
        self.checksum = checksum
        self.checksum_type = checksum_type
        self.version = version
        self._size = size
        self.server = server

    def make_fetcher(self, factory: FetcherFactory) -> FileFetcher:
        return factory.fetch_ngas(self)

    def shares_resources_with(self, other: LocatedFile) -> bool:
        return hasattr(other, "server") and self.server == other.server

    def can_direct_copy(self, context: FetchContext):
        """
        True if this file can be direct copied from NGAS.
        """
        return self.server.can_direct_copy_to(context.path_to(self))

    def direct_copy(self, context: FetchContext):
        """
        Direct copy this file to its destination

        :return: None, if successful
        """
        with self.server.connect() as connection:
            connection.direct_copy(self.ngas_file_id, self.version, context.path_to(self))

    def stream(self, context: FetchContext):
        """
        Stream this file into its destination.

        :return: None, if successful
        """
        with self.server.connect() as connection:
            connection.stream(self.ngas_file_id, self.version, context.path_to(self))

    @property
    def destination(self) -> Path:
        sanitized_subdirectory = self.subdirectory.replace(":", "_").replace("/", "_")
        return Path(sanitized_subdirectory) / Path(self.relative_path)

    @property
    def name(self) -> str:
        return self.relative_path

    @property
    def size(self) -> int:
        return self._size

    @property
    def science_product_locator(self) -> str:
        return self._science_product_locator

    @science_product_locator.setter
    def science_product_locator(self, locator):
        self._science_product_locator = locator

    @property
    def is_sdm(self) -> bool:
        return self.ngas_file_id.endswith(".sdm")

    @property
    def is_image_ancillary(self) -> bool:
        # removing weblog from image ancillaries until processing date determination is sorted
        # return self.relative_path.endswith(".fits") or self.relative_path.endswith("weblog.tgz")
        return self.relative_path.endswith(".fits")

    def __str__(self):
        return f"NGAS {self.ngas_file_id} from {self.server} -> {self.subdirectory}/{self.relative_path}"

    def __eq__(self, other):
        return (
            self.ngas_file_id == other.ngas_file_id
            and self.subdirectory == other.subdirectory
            and self.relative_path == other.relative_path
            and self.checksum == other.checksum
            and self.checksum_type == other.checksum_type
            and self.version == other.version
            and self.size == other.size
            and self.server == other.server
        )


class OracleXml(LocatedFile):
    """Represents the metadata of an ALMA SDM stored as XML in the ALMA DB"""

    archive_uid: str
    table: str
    subdirectory: str
    relative_path: str
    _size: int
    _science_product_locator: str

    def __init__(self, archive_uid: str, table: str, subdirectory: str, relative_path: str, size):
        super().__init__()
        self.archive_uid = archive_uid
        self.table = table
        self.subdirectory = subdirectory
        self.relative_path = relative_path
        self._size = size

    def make_fetcher(self, factory: FetcherFactory) -> FileFetcher:
        return factory.fetch_oracle_xml(self)

    def shares_resources_with(self, other: LocatedFile) -> bool:
        # all OracleXml files use the same resource
        return hasattr(other, "table")

    @property
    def destination(self) -> Path:
        # In the case of ALMA, we'll need to sanitize the UIDs.  Nondestructive for most cases.
        sanitized_subdirectory = self.subdirectory.replace(":", "_").replace("/", "_")
        return Path(sanitized_subdirectory) / Path(self.relative_path)

    def name(self) -> str:
        return self.archive_uid

    @property
    def size(self) -> int:
        return self._size

    @property
    def science_product_locator(self) -> str:
        return self._science_product_locator

    @science_product_locator.setter
    def science_product_locator(self, locator):
        self._science_product_locator = locator

    @property
    def is_sdm(self) -> bool:
        return True

    @property
    def is_image_ancillary(self) -> bool:
        # removing weblog from image ancillaries until processing date determination is sorted
        # return self.relative_path.endswith(".fits") or self.relative_path.endswith("weblog.tgz")
        return self.relative_path.endswith(".fits")

    def __str__(self):
        return f"ALMA {self.archive_uid} @ {self.table} -> {self.subdirectory}/{self.relative_path}"


class FileLocator(Locator):
    """Loads a locations report from a .json report file"""

    def __init__(self, file: Path):
        self.file = file

    def locate(self) -> LocationReport:
        return LocationReportSchema().loads(self.file.read_text())


class ServiceLocator(Locator):
    """Acquires a locations report from the archiveService"""

    def __init__(self, science_product_locator: str):
        self.spl = science_product_locator

    def locate(self) -> LocationReport:
        settings = CapoConfig().settings("edu.nrao.workspaces.ProductFetcherSettings")

        # this is needed to prevent SSL errors when tests are run
        # inside a Docker container
        requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS += ":HIGH:!DH:!aNULL"
        requests.Session().mount(settings.locatorServiceUrlPrefix, adapter=requests.adapters.HTTPAdapter())

        try:
            response = requests.get(settings.locatorServiceUrlPrefix, params={"locator": self.spl})
        except requests.exceptions.Timeout as exc_to:
            raise LocatorServiceException("Locator service timed out") from exc_to
        except requests.exceptions.TooManyRedirects as exc_re:
            raise FetchError("Too many redirects trying to reach locator service") from exc_re
        except requests.exceptions.RequestException as ex:
            raise FetchError(ex.response) from ex

        if response.status_code == http.HTTPStatus.OK:
            return self.append_spl(LocationReportSchema().load(response.json()))
        if response.status_code == http.HTTPStatus.NOT_FOUND:
            raise FetchError(f'locator "{self.spl}" not found')
        raise LocatorServiceException(f"Locator service returned {response.status_code}")

    def append_spl(self, report: LocationReport) -> LocationReport:
        """
        Amend the files in this report with the science product locator we know about.

        :param report:  the report to amend
        :return:  the amended report
        """
        for file in report.files:
            file.science_product_locator = self.spl
        if hasattr(report.ancillaries, "values"):
            for ancillary in report.ancillaries.values():
                self.append_spl(ancillary)
        else:
            for ancillary in report.ancillaries:
                self.append_spl(ancillary)
        return report


class NgasServerSchema(Schema):
    """marshmallow schema to interpret "server" section of a "files" item"""

    server = fields.Str()
    location = fields.Str()
    cluster = fields.Str()

    @post_load
    def make_ngas_server(self, data, **kwargs):
        return NgasServer(data["server"], Location[data["location"]], Cluster[data["cluster"]])


class NgasFileSchema(Schema):
    """One of the items in a location report's "files" list"""

    ngas_file_id = fields.Str()
    subdirectory = fields.Str(allow_none=True)
    relative_path = fields.Str()
    checksum = fields.Integer()
    checksum_type = fields.Str()
    version = fields.Integer()
    size = fields.Integer()
    server = fields.Nested(NgasServerSchema())
    table = fields.Str()
    archive_uid = fields.Str()

    @validates_schema
    def validate_subdirectory(self, data, **kwargs):
        # Fill in a subdirectory value that is None with the default
        if data["subdirectory"] is None:
            data["subdirectory"] = "retrieved-products"

    @post_load
    def make_filespec(self, data, **kwargs):
        if "table" in data:
            return OracleXml(**data)
        return NgasFile(**data)


class SimpleLocationReportSchema(Schema):
    files = fields.List(fields.Nested(NgasFileSchema()))
    aggregate_size = fields.Integer()

    @post_load
    def make_report(self, data, **kwargs):
        return LocationReport(**data)


class LocationReportSchema(Schema):
    """Encapsulates an entire locations report"""

    files = fields.List(fields.Nested(NgasFileSchema()))
    aggregate_size = fields.Integer()
    ancillaries = fields.Dict(keys=fields.String(), values=fields.Nested(SimpleLocationReportSchema()))

    @post_load
    def make_report(self, data, **kwargs):
        return LocationReport(**data)
